/* pam_usertty module */

#define USERTTY_FILE	"/etc/usertty"
#define TTY_PREFIX     "/dev/"

/*
 * by Pablo Saratxaga <srtxg@chanae.alphanet.ch>
 * June 18, 1998.
 * - Heavily Based on pam_securetty 
 *   by Elliot Lee <sopwith@redhat.com>, Red Hat Software.
 *   July 25, 1996 which was based on pam_rootok module.
 * - The actual checking of /etc/usertty taken from checktty.c
 *   of util-linux, author Peter Orbaek <poe@daimi.aau.dk>
 * - am_in_group() by Tchi Southivong <tchi@eden.frmug.org> 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <syslog.h>
#include <stdarg.h>
#include <pwd.h>
#include <strings.h>
#include <sys/param.h>
#include <netdb.h>
#include <time.h>
#include <grp.h>

#ifdef __linux__
#  include <linux/major.h>
#  include <linux/kdev_t.h>
#endif

/*
 * here, we make a definition for the externally accessible function
 * in this file (this definition is required for static a module
 * but strongly encouraged generally) it is used to instruct the
 * modules include file to define the function prototypes.
 */

#define PAM_SM_AUTH

#include <security/pam_modules.h>

/* some syslogging */
/* closelog first to make sure it'll _really_ go to syslog */

static void _pam_log(int err, const char *format, ...)
{
    va_list args;

    va_start(args, format);
    closelog();
    openlog("PAM-usertty", LOG_CONS|LOG_PID, LOG_AUTH);
    vsyslog(err, format, args);
    va_end(args);
    closelog();
}

/* argument parsing */

#define PAM_DEBUG_ARG       0x0001

static int _pam_parse(int argc, const char **argv)
{
    int ctrl=0;

    /* step through arguments */
    for (ctrl=0; argc-- > 0; ++argv) {

	/* generic options */

	if (!strcmp(*argv,"debug"))
	    ctrl |= PAM_DEBUG_ARG;
	else {
	    _pam_log(LOG_ERR,"pam_parse: unknown option; %s",*argv);
	}
    }

    return ctrl;
}

/* --- various struct and functions from checktty.c --- */

#define NAMELEN 128

/* linked list of names */
struct grplist {
    struct grplist *next;
    char name[NAMELEN];
};
        
enum State { StateUsers, StateGroups, StateClasses };

#define CLASSNAMELEN 32

struct ttyclass {
    struct grplist *first;
    struct ttyclass *next;
    char classname[CLASSNAMELEN];
};

struct ttyclass *ttyclasses = NULL;

static int
am_in_group(struct passwd *p,char *group)
{
    struct group *g;
    int i=0;
    char *s;

    if ( (g=getgrnam(group)) ) {
        if (p->pw_gid==g->gr_gid)
	    return(1);
        while ( (s=(g->gr_mem)[i++]) )
	    if (!strncmp(p->pw_name,s,8))
		return(1);
    }
    return(0);
}

static struct ttyclass *
new_class(char *class)
{
    struct ttyclass *tc;

    tc = (struct ttyclass *)malloc(sizeof(struct ttyclass));
    if (tc == NULL) {
	printf("login: memory low, login may fail\n");
	syslog(LOG_WARNING, "can't malloc for ttyclass");
	return NULL;
    }

    tc->next = ttyclasses;
    tc->first = NULL;
    strncpy(tc->classname, class, CLASSNAMELEN);
    tc->classname[CLASSNAMELEN-1] = 0;
    ttyclasses = tc;
    return tc;
}

static void
add_to_class(struct ttyclass *tc, char *tty)
{
    struct grplist *ge;

    if (tc == NULL) return;

    ge = (struct grplist *)malloc(sizeof(struct grplist));
    if (ge == NULL) {
	printf("login: memory low, login may fail\n");
	syslog(LOG_WARNING, "can't malloc for grplist");
	return;
    }

    ge->next = tc->first;
    strncpy(ge->name, tty, NAMELEN);
    ge->name[NAMELEN-1] = 0;
    tc->first = ge;
}


/* return true if tty is a pty. Very linux dependent */
static int
isapty(const char *tty)
{
    char devname[100];
    struct stat stb;

    snprintf(devname, sizeof(devname), "/dev/%s", tty);

#if defined(__linux__) && defined(PTY_SLAVE_MAJOR)

    /* this is for linux 1.3 and newer */
    if((stat(devname, &stb) >= 0)
	&& MAJOR(stb.st_rdev) == PTY_SLAVE_MAJOR) {
	return 1;
    }
#endif

    return 0;
}

/* match the hostname hn against the pattern pat */
static int
hnmatch(const char *hn, const char *pat)
{
    int x1, x2, x3, x4, y1, y2, y3, y4;
    unsigned long p, mask, a;
    unsigned char *ha;
    struct hostent *he;
    int n, m;

    if ((hn == NULL) && (strcmp(pat, "localhost") == 0)) return 1;
    if ((hn == NULL) || hn[0] == 0) return 0;

    if (pat[0] >= '0' && pat[0] <= '9') {
	/* pattern is an IP QUAD address and a mask x.x.x.x/y.y.y.y */
	sscanf(pat, "%d.%d.%d.%d/%d.%d.%d.%d", &x1, &x2, &x3, &x4,
	       &y1, &y2, &y3, &y4);
	p = (((unsigned long)x1<<24)+((unsigned long)x2<<16)
	     +((unsigned long)x3<<8)+((unsigned long)x4));
	mask = (((unsigned long)y1<<24)+((unsigned long)y2<<16)
		+((unsigned long)y3<<8)+((unsigned long)y4));

	he = gethostbyname(hn);
	if (!he || !he->h_addr_list || !he->h_addr_list[0])
	  return 0;

	ha = (unsigned char *)he->h_addr_list[0];
	a = (((unsigned long)ha[0]<<24)+((unsigned long)ha[1]<<16)
	     +((unsigned long)ha[2]<<8)+((unsigned long)ha[3]));
	return ((p & mask) == (a & mask));
    } else {
	/* pattern is a suffix of a FQDN */
	n = strlen(pat);
	m = strlen(hn);
	if (n > m) return 0;
	return (strcasecmp(pat, hn + m - n) == 0);
    }
}

const static char *wdays[] = { "sun", "mon", "tue", "wed", "thu", "fri", "sat" };

/* example timespecs:

   mon:tue:wed:8-17

   meaning monday, tuesday or wednesday between 8:00 and 17:59

   4:5:13:fri

   meaning fridays from 4:00 to 5:59 and from 13:00 to 13:59
*/
static int
timeok(struct tm *t, char *spec)
{
    char *p, *q;
    int dayok = 0;
    int hourok = 0;
    int h, h2;
    char *sp;

    sp = spec;
    while ((p = strsep(&sp, ":"))) {
	if (*p >= '0' && *p <= '9') {
	    h = atoi(p);
	    if (h == t->tm_hour) hourok = 1;
	    if ((q = strchr(p, '-')) && (q[1] >= '0' && q[1] <= '9')) {
		h2 = atoi(q+1);
		if (h <= t->tm_hour && t->tm_hour <= h2) hourok = 1;
	    }
	} else if (strcasecmp(wdays[t->tm_wday], p) == 0) {
	    dayok = 1;
	}
    }

    return (dayok && hourok);
}

/* return true if tty equals class or is in the class defined by class.
   Also return true if hostname matches the hostname pattern, class
   or a pattern in the class named by class. */
static int
in_class(const char *tty, const char *hostname, char *class)
{
    struct ttyclass *tc;
    struct grplist *ge;
    time_t t;
    char *p;
    char timespec[256];
    struct tm *tm;
    char *n;

    time(&t);
    tm = localtime(&t);

    if (class[0] == '[') {
	if ((p = strchr(class, ']'))) {
	    *p = 0;
	    strncpy(timespec, class+1, sizeof(timespec));
	    timespec[sizeof(timespec)-1] = 0;
	    *p = ']';
	    if(!timeok(tm, timespec)) return 0;
	    class = p+1;
	}
	/* really ought to warn about syntax error */
    }

    if (strcmp(tty, class) == 0) return 1;

    if ((class[0] == '@') && isapty(tty)
	&& hnmatch(hostname, class+1)) return 1;

    for (tc = ttyclasses; tc; tc = tc->next) {
	if (strcmp(tc->classname, class) == 0) {
	    for (ge = tc->first; ge; ge = ge->next) {

		n = ge->name;
		if (n[0] == '[') {
		    if ((p = strchr(n, ']'))) {
			*p = 0;
			strncpy(timespec, n+1, sizeof(timespec));
			timespec[sizeof(timespec)-1] = 0;
			*p = ']';
			if(!timeok(tm, timespec)) continue;
			n = p+1;
		    }
		    /* really ought to warn about syntax error */
		}

		if (strcmp(n, tty) == 0) return 1;

		if ((n[0] == '@') && isapty(tty)
		    && hnmatch(hostname, n+1)) return 1;
	    }
	    return 0;
	}
    }
    return 0;
}

/* start JDS - SBA */

void free_group(struct grplist *ge);
void free_class(struct ttyclass *tc);
void free_all(void);

void 
free_group(struct grplist *ge)
{
    if (ge) {
	memset(ge->name, 0, NAMELEN);
	free_group(ge->next);
	free(ge->next);
	ge->next = NULL;
    }
}

void 
free_class(struct ttyclass *tc)
{
    if (tc) {
	memset(tc->classname, 0, CLASSNAMELEN);
	free_group(tc->first);
	tc->first = NULL;
	free_class(tc->next);
	free(tc->next);
	tc->next = NULL;
    }
}

void 
free_all(void)
{
    free_class(ttyclasses);
    ttyclasses = NULL;
}
/* end JDS - SBA */


/* --- authentication management functions (only) --- */

PAM_EXTERN
int pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc
			,const char **argv)
{
    int retval = PAM_AUTH_ERR;
    const char *username, *hostname;
    char *uttyname;
    char buf[256], defaultbuf[256];
    enum State state = StateUsers;
    int found_match = 0;
    struct passwd *user_pwd;
    struct stat userttyfileinfo;
    FILE *userttyfile;
    int ctrl;
    char *p;

    /* parse the arguments */
    ctrl = _pam_parse(argc, argv);

    retval = pam_get_item(pamh,PAM_USER,(const void **)&username);
    if (retval == PAM_SUCCESS) {
	pam_get_item(pamh,PAM_TTY,(const void **)&uttyname);
	pam_get_item(pamh,PAM_RHOST,(const void **)&hostname);
    }
    if (retval != PAM_SUCCESS || (uttyname == NULL && hostname == NULL)) {
	/* If we couldn't get the username, the tty nor the hostname
           return error */
        if (ctrl & PAM_DEBUG_ARG)
            _pam_log(LOG_WARNING, "can not determine tty I'm running on !");
	return PAM_SERVICE_ERR;
    }

    /* The PAM_TTY item may be prefixed with "/dev/" - skip that */
    if (strncmp(TTY_PREFIX, uttyname, sizeof(TTY_PREFIX)-1) == 0)
	uttyname += sizeof(TTY_PREFIX)-1;

    /* If we didn't get a username, get one */
    if(!username || (strlen(username) <= 0)) {
	/* Don't let them use a NULL username... */
	(void) pam_set_item(pamh, PAM_USER, NULL);
	pam_get_user(pamh,&username,NULL);
	if (retval != PAM_SUCCESS || username == NULL || *username == '\0') {
	    if (ctrl & PAM_DEBUG_ARG)
		_pam_log(LOG_WARNING, 
			 "can not determine username for this service!");
	    return PAM_SERVICE_ERR;
	}
    }

    user_pwd = getpwnam(username);
    if (user_pwd == NULL)
        return PAM_IGNORE;

    if(stat(USERTTY_FILE,&userttyfileinfo)) {
	_pam_log(LOG_NOTICE,
		 "Couldn't open " USERTTY_FILE);
	return PAM_SUCCESS; /* for compatibility with old securetty handling,
			       this needs to succeed.  But we still log the
			       error. */
    }

    if((userttyfileinfo.st_mode & S_IWOTH)
       || !S_ISREG(userttyfileinfo.st_mode)) {
	/* If the file is world writable or is not a
	   normal file, return error */
	_pam_log(LOG_ERR, USERTTY_FILE
		 " is either world writable or not a normal file");
	return PAM_AUTH_ERR;
    }

    userttyfile = fopen(USERTTY_FILE,"r");
    if(userttyfile == NULL) { /* Check that we opened it successfully */
	_pam_log(LOG_ERR,
		 "Error opening " USERTTY_FILE);
	return PAM_SERVICE_ERR;
    }

    /* There should be no more errors from here on */
    retval=PAM_AUTH_ERR;

    defaultbuf[0] = 0;
    while(fgets(buf, 255, userttyfile)) {

	buf[255] = '\0';

	/* strip comments */
	for (p = buf; p < buf + 256; p++) {
            if (*p == '#') {
                *p = '\0';
                break; /* ignore the rest of the line */
            }
        }

	if (buf[0] == '*') {
	    strncpy(defaultbuf, buf, 256);
	    defaultbuf[255] = 0;
	    continue;
	}

	if (strncmp("GROUPS", buf, 6) == 0) {
	    state = StateGroups;
	    continue;
	} else if (strncmp("USERS", buf, 5) == 0) {
	    state = StateUsers;
	    continue;
	} else if (strncmp("CLASSES", buf, 7) == 0) {
	    state = StateClasses;
	    continue;
	}

	strtok(buf, " \t");
	if((state == StateUsers && (strncmp(username, buf, 8) == 0))
	    || (state == StateGroups && am_in_group(user_pwd, buf))) {
	    found_match = 1;  /* we found a line matching the username */
	    while((p = strtok(NULL, "\t\n "))) {
		if (in_class(uttyname, hostname, p)) {
		    retval = PAM_SUCCESS;
		    break;
		}
	    }
	} else if (state == StateClasses) {
	    /* define a new tty/host class */
	    struct ttyclass *tc = new_class(buf);
	    while ((p = strtok(NULL, "\t\n "))) {
		add_to_class(tc, p);
	    }
	}
    }
    fclose(userttyfile);

    /* username is not explicitly mentioned in /etc/usertty, if there was
       a default rule, use that */
    if ((!found_match) && (defaultbuf[0])) {
	found_match = 1;
	strtok(defaultbuf, " \t");
	while((p = strtok(NULL, "\t\n "))) {
	    if (in_class(uttyname, hostname, p)) {
		retval = PAM_SUCCESS;
	    }
	}
    }

    /* users not matched in /etc/usertty are by default allowed access
       on all tty's */
    if (!found_match) {
	retval = PAM_SUCCESS;
    }

    /* free various structures used */
    free_all();

    if(retval) {
	if (ctrl & PAM_DEBUG_ARG)
	    _pam_log(LOG_WARNING,
		     "access denied: user '%s' can't use tty '%s' from %s",
		     username,
		     uttyname,
		     hostname);
	retval = PAM_AUTH_ERR;
    }
    if ((retval == PAM_SUCCESS) && (ctrl & PAM_DEBUG_ARG))
	_pam_log(LOG_DEBUG, "access allowed for '%s' on '%s'",
		 username, uttyname);
    return retval;

}

PAM_EXTERN
int pam_sm_setcred(pam_handle_t *pamh,int flags,int argc
		   ,const char **argv)
{
     return PAM_SUCCESS;
}


#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_usertty_modstruct = {
     "pam_usertty",
     pam_sm_authenticate,
     pam_sm_setcred,
     NULL,
     NULL,
     NULL,
     NULL,
};

#endif

/* end of module definition */
